#!/usr/bin/env python
# getAllPossibleSpliceJunctions.py
# Author: Angela Brooks
# Program Completion Date:
# Description: Joins all possible junctions.  Only creates a splice site if at
# least one half-site is within an annotated gene.  Also creates novel
# combinations of junctions between all exons within a gene and creates reverse
# splicing junctions.
# [A,R,KK, NK, KN, NN]  chr strand intron_start intron_end 
# A - Annotated
# G - An unannotated juction that comes from two splice sites of the same gene.
# R - An unannotated junction from splice sites that come from an annotated 5'
# splice site with any upstream 3' splice site.  It is essentially like a
# reverse splicing.
# KK - Two known splice sites joined together to make a novel junction, but the
# splice sites come from different gene loci.
# NK - 5' splice site is novel, 3' splice site is known
# KN - 5' splice site is known, 3' splice site is novel
# NN - Two novel splice sites joined together.
# Modification Date(s):

import sys
import optparse 
import time
import pdb

from Bio import SeqIO
from Bio.Seq import Seq

from sequence_wrap import DNA
from helperFunctions import updateDictOfSets
#############
# CONSTANTS #
#############
DEF_NOVEL_MIN_INTRON_LEN = 20
DEF_NOVEL_MAX_INTRON_LEN = 2000

# Splice site variables
EXONIC_SSS_LEN = 3
FIVEP_SSS_INTRONIC_LEN = 6
THREEP_SSS_INTRONIC_LEN = 20

TRUE = 1
FALSE = 0

HOST = "localhost"
USER = "root"
PASSWD = ""

#VALID_CHR = ["2L", "2LHet", "2R", "2RHet", "3L", "3LHet",
#             "3R", "3RHet", "4", "X", "XHet", "YHet", "U"]

#VALID_CHR = "2L,2LHet,2R,2RHet,3L,3LHet,3R,3RHet,4,X,XHet,YHet,U"
#VALID_CHR = ["4"]

# Had problems with header delimeter given certain gene names, so new delimiter
# is give here.
DELIM = "|"
#AS_DB_APPENDED_SYMB = "__"
#################
# END CONSTANTS #
#################


###########
# CLASSES #
###########
class OptionParser(optparse.OptionParser):
    """
    Adding a method for required arguments.
    Taken from:
    http://www.python.org/doc/2.3/lib/optparse-extending-examples.html
    """
    def check_required(self, opt):
        option = self.get_option(opt)

        # Assumes the option's 'default' is set to None!
        if getattr(self.values, option.dest) is None:
            print "%s option not supplied" % option
            self.print_help()
            sys.exit(1)


###############
# END CLASSES #
###############
 
########
# MAIN #	
########
def main():
	
    opt_parser = OptionParser()
   
    # Add Options. Required options should have default=None
    opt_parser.add_option("-t",
                          dest="txt_db",
                          type="string",
                          help="Transcript database name",
                          default=None)
    opt_parser.add_option("--sqlite_db_dir",
                          dest="sqlite_db_dir",
                          type="string",
                          help="""Location of sqlite database. Will override
                                  usage of MySQL database.""",
                          default=None)
    opt_parser.add_option("--host",
                          dest="host",
                          type="string",                          
                          help="MySQL database host. Default=%s" % HOST,
                          default=HOST)
    opt_parser.add_option("--user",
                          dest="user",
                          type="string",                          
                          help="MySQL database user. Default=%s" % USER,
                          default=USER)
    opt_parser.add_option("--passwd",
                          dest="passwd",
                          type="string",                          
                          help="MySQL database passwd. Default=%s" % PASSWD,
                          default=PASSWD)
    opt_parser.add_option("--chr",
                          dest="chr_file",
                          type="string",                          
                          help="""File containing chromosome sequences from the
                                  genome.""",
                          default=None)
    opt_parser.add_option("--novel",
                          dest="novel_flag",
                          type="string",
                          help="""\'all\' indicates two novel splice site
                                  junctions(NN) will be given. \'noNN\'
                                  indicates novel splice sites included, but
                                  only if joined with a known splice sites
                                  (KN,NK).  No option given means no novel
                                  splice sites are given.""",
                          default=None)
    opt_parser.add_option("--min",
                          dest="min",
                          type="int",
                          help="Minimum length of a novel junction. Def=%d" %\
                                DEF_NOVEL_MIN_INTRON_LEN,
                          default=DEF_NOVEL_MIN_INTRON_LEN)
    opt_parser.add_option("--max",
                          dest="max",
                          type="int",
                          help="Maximum length of a novel junction. Def=%d" %\
                                DEF_NOVEL_MAX_INTRON_LEN,
                          default=DEF_NOVEL_MAX_INTRON_LEN)
    opt_parser.add_option("--print_fasta",
                          dest="overhang",
                          type="int",
                          help="""Will also print out all the junctions
                                  sequences with the given overhang length on
                                  both sides.""",
                          default=None)
    opt_parser.add_option("--rev_comp",
                          dest="rev_comp",
                          action="store_true",
                          help="""If this option is given, the sequence of
                                  junctions on the minus strand will be given
                                  as the reverse complement of the genome.
                                  Default: Sequence is always given in the
                                  Watson(+) orientation.""",
                          default=False)
    opt_parser.add_option("--use_unknown_strand",
                          dest="use_unknown_strand",
                          action="store_true",
                          help="""By default, annotated junctions with unknown
                                  strands ('.') will not be considered.  If
                                  this option is given, then the splice sites
                                  for these junctions will be considered in 
                                  both the plus and the minus orientations. The
                                  annotated version will only be reported in
                                  the '+' orientation but when looking for
                                  additional unannotated junctions, it will be combined
                                  with splice sites on both strands. """,
                          default=False)
    opt_parser.add_option("-o",
                          dest="output",
                          type="string",
                          help="""Output file that will contain junction
                                  coordinates.""",
                          default=None)
    opt_parser.add_option("-p",
                          dest="prefix",
                          type="string",
                          help="""Prefix that will be added to all output
                                  files.""",
                          default=None)

    (options, args) = opt_parser.parse_args()
	
    # validate the command line arguments
    opt_parser.check_required("-t")
    opt_parser.check_required("--chr")
    opt_parser.check_required("-o")
    opt_parser.check_required("-p")
    
    txt_db = options.txt_db
    min_len = options.min
    max_len = options.max
    chr_file = open(options.chr_file)

    novel = options.novel_flag
    
    if novel:
        if novel != "all" and novel != "noNN":
            print "Must specify \'all\' or \'noNN\' for --novel option"
            opt_parser.print_help()
            sys.exit(1)

    overhang = options.overhang
    rev_comp = options.rev_comp
    use_unknown_strand = options.use_unknown_strand
     
    prefix = options.prefix

    outfile_name = options.output

    db = None
    if options.sqlite_db_dir:
        import pysqlite_wrap
        db = pysqlite_wrap.DB(options.sqlite_db_dir)
    else: # Use MySQL database
        import mysqldb_wrap
        db = mysqldb_wrap.DB(options.host, options.user, options.passwd)


    print_fasta = False
    if overhang:
        print_fasta = True

    only_known = True
    if novel:
        only_known = False

    # If you want to print a fasta file, then you need a output fasta for each
    # type of junction
    if print_fasta:
        a_file = open("%s_A_junctions.fa" % prefix, "w")
        g_file = open("%s_G_junctions.fa" % prefix, "w")
        r_file = open("%s_R_junctions.fa" % prefix, "w")
        kk_file = open("%s_KK_junctions.fa" % prefix, "w")
       
        nk_file = None
        kn_file = None
        nn_file = None 
        if novel:
            nk_file = open("%s_NK_junctions.fa" % prefix, "w")
            kn_file = open("%s_KN_junctions.fa" % prefix, "w")
            if novel == "all":
                nn_file = open("%s_NN_junctions.fa" % prefix, "w")

    outfile = open(outfile_name, "w")
    
    
    for record in SeqIO.parse(chr_file, "fasta"):

        chr = record.id 

        print "Working on %s" % chr

        # If you want to print a fasta file
        # Then you need to open the sequence file in the chromosome directory
        if print_fasta:
            chr_seq = DNA(record.seq.tostring())
        

        # joined_set_plus and joined_set_minus are used to make sure duplicate
        # junctions are not produced later
        if print_fasta:
            (joined_set_plus, 
             joined_set_minus,
             annotated_fivep_ss_plus_pos2gene,
             annotated_fivep_ss_minus_pos2gene,
             annotated_threep_ss_plus_pos2gene,
             annotated_threep_ss_minus_pos2gene) = printJcnsFromAnnotatedSS(db, txt_db, chr,
                                                          chr_seq, 
                                                          outfile,
                                                          rev_comp,
                                                          use_unknown_strand,
                                                          a_file,
                                                          g_file,
                                                          r_file,
                                                          overhang)
        else:
            (joined_set_plus, 
             joined_set_minus,
             annotated_fivep_ss_plus_pos2gene,
             annotated_fivep_ss_minus_pos2gene,
             annotated_threep_ss_plus_pos2gene,
             annotated_threep_ss_minus_pos2gene)  = printJcnsFromAnnotatedSS(db, txt_db, chr,
                                                          chr_seq, 
                                                          outfile,
                                                          rev_comp,
                                                          use_unknown_strand)

        if len(joined_set_plus) == 0 and len(joined_set_minus) == 0:
            print "No junctions for %s" % chr
            continue

        # Returns sets of genes that overlap each position in the chromosome
        print "Finding Gene Positions"
        t0 = time.time()
        pos2gene_set_plus = getGenePositions(db, chr, "+", txt_db)
        pos2gene_set_minus = getGenePositions(db, chr, "-", txt_db)
        print time.time() - t0, "seconds"


        # Now find all splice sites and their associated genes
        print "Finding All Splice Sites"
        t0 = time.time()

        (fivep_plus_pos_inGene, fivep_minus_pos_inGene,
         fivep_plus_pos_not_inGene, fivep_minus_pos_not_inGene,
         threep_plus_pos_inGene, threep_minus_pos_inGene,
         threep_plus_pos_not_inGene, threep_minus_pos_not_inGene) = getAllSpliceSites(db, chr, chr_seq,
                                                                pos2gene_set_plus,
                                                                pos2gene_set_minus,
                                                                annotated_fivep_ss_plus_pos2gene,
                                                                annotated_fivep_ss_minus_pos2gene,
                                                                annotated_threep_ss_plus_pos2gene,
                                                                annotated_threep_ss_minus_pos2gene,
                                                                only_known)
    
        print time.time() - t0, "seconds"

        # Sort lists
        fivep_plus_pos_inGene.sort()
        fivep_minus_pos_inGene.sort()
        threep_plus_pos_inGene.sort()
        threep_minus_pos_inGene.sort()

        fivep_plus_pos_not_inGene.sort()
        fivep_minus_pos_not_inGene.sort()
        threep_plus_pos_not_inGene.sort()
        threep_minus_pos_not_inGene.sort()
    

        print "Combining inGene:inGene splice sites '+'"
        t0 = time.time()        
        combineSpliceSites(fivep_plus_pos_inGene, threep_plus_pos_inGene, "+",
                           joined_set_plus,
                           annotated_fivep_ss_plus_pos2gene, 
                           annotated_threep_ss_plus_pos2gene,
                           pos2gene_set_plus, 
                           print_fasta,
                           outfile, kk_file, kn_file, nk_file, nn_file,
                           chr, chr_seq, overhang, min_len, max_len, rev_comp)
        print time.time() - t0, "seconds"

        print "Combining inGene:inGene splice sites '-'"
        t0 = time.time()        
        combineSpliceSites(threep_minus_pos_inGene, fivep_minus_pos_inGene, "-",
                           joined_set_minus,
                           annotated_threep_ss_minus_pos2gene, 
                           annotated_fivep_ss_minus_pos2gene,
                           pos2gene_set_minus, 
                           print_fasta,
                           outfile, kk_file, kn_file, nk_file, nn_file,
                           chr, chr_seq, overhang, min_len, max_len, rev_comp)
        print time.time() - t0, "seconds"

        print "Combining inGene:not_inGene splice sites '+'"
        t0 = time.time()        
        combineSpliceSites(fivep_plus_pos_inGene, threep_plus_pos_not_inGene, "+",
                           joined_set_plus,
                           annotated_fivep_ss_plus_pos2gene, 
                           annotated_threep_ss_plus_pos2gene,
                           pos2gene_set_plus, 
                           print_fasta,
                           outfile, kk_file, kn_file, nk_file, nn_file,
                           chr, chr_seq, overhang, min_len, max_len, rev_comp)
        print time.time() - t0, "seconds"

        print "Combining inGene:not_inGene splice sites '-'"
        t0 = time.time()        
        combineSpliceSites(threep_minus_pos_inGene, fivep_minus_pos_not_inGene, "-",
                           joined_set_minus,
                           annotated_threep_ss_minus_pos2gene, 
                           annotated_fivep_ss_minus_pos2gene,
                           pos2gene_set_minus, 
                           print_fasta,
                           outfile, kk_file, kn_file, nk_file, nn_file,
                           chr, chr_seq, overhang, min_len, max_len, rev_comp)
        print time.time() - t0, "seconds"


        print "Combining not_inGene:inGene splice sites '+'"
        t0 = time.time()        
        combineSpliceSites(fivep_plus_pos_not_inGene, threep_plus_pos_inGene, "+",
                           joined_set_plus,
                           annotated_fivep_ss_plus_pos2gene, 
                           annotated_threep_ss_plus_pos2gene,
                           pos2gene_set_plus, 
                           print_fasta,
                           outfile, kk_file, kn_file, nk_file, nn_file,
                           chr, chr_seq, overhang, min_len, max_len, rev_comp)
        print time.time() - t0, "seconds"


        print "Combining not_inGene:inGene splice sites '-'"
        t0 = time.time()        
        combineSpliceSites(threep_minus_pos_not_inGene, fivep_minus_pos_inGene, "-",
                           joined_set_minus,
                           annotated_threep_ss_minus_pos2gene, 
                           annotated_fivep_ss_minus_pos2gene,
                           pos2gene_set_minus, 
                           print_fasta,
                           outfile, kk_file, kn_file, nk_file, nn_file,
                           chr, chr_seq, overhang, min_len, max_len, rev_comp)
        print time.time() - t0, "seconds"


    sys.exit(0)

############
# END_MAIN #
############

#############
# FUNCTIONS #
#############
def formatLine(line):
    line = line.replace("\r","")
    line = line.replace("\n","")
    return line

def combineSpliceSites(upstrm_pos_list, dwnstrm_pos_list, strand, joined_set,
                       annotated_upstrm_pos2gene, annotated_dwnstrm_pos2gene,
                       pos2gene_set, 
                       print_fasta,
                       outfile, kk_file, kn_file, nk_file, nn_file,
                       chr, chr_seq,
                       overhang,
                       min_len, max_len, rev_comp):

    dwnstrm_idx_starter = 0
    if len(upstrm_pos_list) == 0 or len(dwnstrm_pos_list) == 0:
        return
    
    if upstrm_pos_list == [] or dwnstrm_pos_list == []:
        return

    for upstrm_idx in range(len(upstrm_pos_list)):
        upstrm_pos = upstrm_pos_list[upstrm_idx]

        while dwnstrm_pos_list[dwnstrm_idx_starter] <= upstrm_pos: 
            dwnstrm_idx_starter += 1

            if dwnstrm_idx_starter == len(dwnstrm_pos_list):
                break

        if dwnstrm_idx_starter == len(dwnstrm_pos_list):
            break

        # Check for acceptable dwnstrm_pos
        for dwnstrm_idx in range(dwnstrm_idx_starter,len(dwnstrm_pos_list)):
            dwnstrm_pos = dwnstrm_pos_list[dwnstrm_idx]

            # Either the fivep or threep has to be in a gene region.  No
            # completely intergenic junctions
#                if not (fivep_pos in pos2gene_set_plus or 
#                        threep_pos in pos2gene_set_plus):
#                    continue
            intron_len = dwnstrm_pos - upstrm_pos + 1

            if intron_len < min_len:
                continue
       
            if intron_len <= max_len:
                # Found a junction, but need to consider the type
                # Check if it already exists in the joined set
                if (upstrm_pos, dwnstrm_pos) in joined_set:
                    continue

                if ((upstrm_pos in annotated_upstrm_pos2gene) and 
                    (dwnstrm_pos in annotated_dwnstrm_pos2gene)):
                    type = "KK"
                    type_file = kk_file

                    upstrm_genes = ",".join(list(annotated_upstrm_pos2gene[upstrm_pos]))
                    dwnstrm_genes = ",".join(list(annotated_dwnstrm_pos2gene[dwnstrm_pos]))
                        

                elif upstrm_pos in annotated_upstrm_pos2gene:
                    if strand == "+":
                        type = "KN"
                        type_file = kn_file
                    else:
                        type = "NK"
                        type_file = nk_file

                    upstrm_genes = ",".join(list(annotated_upstrm_pos2gene[upstrm_pos]))
                    dwnstrm_genes = "None"
                    if dwnstrm_pos in pos2gene_set:
                        dwnstrm_genes = ",".join(list(pos2gene_set[dwnstrm_pos]))


                elif dwnstrm_pos in annotated_dwnstrm_pos2gene:
                    if strand == "+":
                        type = "NK"
                        type_file = nk_file
                    else:
                        type = "KN"
                        type_file = kn_file

                    upstrm_genes = "None"
                    if upstrm_pos in pos2gene_set:
                        upstrm_genes = ",".join(list(pos2gene_set[upstrm_pos]))
                    dwnstrm_genes = ",".join(list(annotated_dwnstrm_pos2gene[dwnstrm_pos]))

                else:
                    # Do not consider NN junctions when the nn_file does not
                    # exist
                    if not nn_file:
                        continue

                    type = "NN"
                    type_file = nn_file

                    upstrm_genes = "None"
                    if upstrm_pos in pos2gene_set:
                        upstrm_genes = ",".join(list(pos2gene_set[upstrm_pos]))    
                    dwnstrm_genes = "None"
                    if dwnstrm_pos in pos2gene_set:
                        dwnstrm_genes = ",".join(list(pos2gene_set[dwnstrm_pos]))

                if strand == "+":
                    fivep_genes = upstrm_genes
                    threep_genes = dwnstrm_genes
                else:
                    fivep_genes = dwnstrm_genes
                    threep_genes = upstrm_genes
                

                # Print junction
                print_str = "%s\t%s%s%s\t%s\t%s\t%d\t%d\n" % (type, 
                                                            fivep_genes,
                                                            DELIM,
                                                            threep_genes,
                                                            chr,
                                                            strand,
                                                            upstrm_pos,
                                                            dwnstrm_pos)

                outfile.write(print_str)
                if print_fasta:
                    writeToFasta(type_file, print_str, chr_seq, upstrm_pos,
                                 dwnstrm_pos, strand, overhang, rev_comp)
        

            else: # No more threep ss to check for this fivep ss
                break





def formatChr(chr):
    if not chr.startswith("chr"):
        return "chr" + chr
    return chr

def getAllSpliceSites(db, chr, chr_seq, 
                      pos2gene_set_plus,
                      pos2gene_set_minus,
                      annotated_fivep_ss_plus_pos2gene,
                      annotated_fivep_ss_minus_pos2gene,
                      annotated_threep_ss_plus_pos2gene,
                      annotated_threep_ss_minus_pos2gene,    
                      only_known):
    """
    Returns:
        fivep_plus_pos, fivep_minus_pos,
        threep_plus_pos, threep_minus_pos,
    """
    # Return lists [pos,]
    fivep_plus_pos_inGene = list(annotated_fivep_ss_plus_pos2gene.keys())
    fivep_minus_pos_inGene = list(annotated_fivep_ss_minus_pos2gene.keys())
    threep_plus_pos_inGene = list(annotated_threep_ss_plus_pos2gene.keys())
    threep_minus_pos_inGene = list(annotated_threep_ss_minus_pos2gene.keys())

    fivep_plus_pos_not_inGene = []
    fivep_minus_pos_not_inGene = []
    threep_plus_pos_not_inGene = []
    threep_minus_pos_not_inGene = []

    if only_known:
        return (fivep_plus_pos_inGene, fivep_minus_pos_inGene, 
                fivep_plus_pos_not_inGene, fivep_minus_pos_not_inGene, 
                threep_plus_pos_inGene, threep_minus_pos_inGene,
                threep_plus_pos_not_inGene, threep_minus_pos_not_inGene)
    

    chr_length = len(chr_seq)

    # Although the chr_seq is indexed from 0, the reported coordinates are 1
    # based.
    for i in range(chr_length):
        # check 5' splice sites on forward strand
        if chr_seq[i:i+2].tostring() == "GT":
            pos = i+1
            if pos in annotated_fivep_ss_plus_pos2gene:
                continue
            # Check length constraints
            if EXONIC_SSS_LEN <= i <= (chr_length - FIVEP_SSS_INTRONIC_LEN):
                # Found a splice site
                if pos in pos2gene_set_plus:
                    fivep_plus_pos_inGene.append(pos)            
                else:
                    fivep_plus_pos_not_inGene.append(pos)

        # Check 5' splice sites on the reverse strand
        elif chr_seq[i:i+2].tostring() == "AC":
            pos = i+2
            if pos in annotated_fivep_ss_minus_pos2gene:
                continue
            # Check length constraints
            if (FIVEP_SSS_INTRONIC_LEN - 2) <= i <= (chr_length - EXONIC_SSS_LEN - 2):
                # Found a splice site
                if pos in pos2gene_set_minus:
                    fivep_minus_pos_inGene.append(pos) 
                else:
                    fivep_minus_pos_not_inGene.append(pos)

        # Check 3' splice site on forward strand
        elif chr_seq[i:i+2].tostring() == "AG":
            pos = i+2
            if pos in annotated_threep_ss_plus_pos2gene:
                continue
            # Check length constraints
            if (THREEP_SSS_INTRONIC_LEN - 2) <= i <= (chr_length - EXONIC_SSS_LEN - 2):
                # Found a splice site
                if pos in pos2gene_set_plus:
                    threep_plus_pos_inGene.append(pos)
                else:
                    threep_plus_pos_not_inGene.append(pos)

        # Check 3' splice site on reverse strand
        elif chr_seq[i:i+2].tostring() == "CT":
            pos = i+1
            if pos in annotated_threep_ss_minus_pos2gene:
                continue
            # Check length constraints
            if EXONIC_SSS_LEN <= i <= (chr_length - THREEP_SSS_INTRONIC_LEN):
                # Found a splice site
                if pos in pos2gene_set_minus:
                    threep_minus_pos_inGene.append(pos)
                else:
                    threep_minus_pos_not_inGene.append(pos)

    return (fivep_plus_pos_inGene, fivep_minus_pos_inGene, 
            fivep_plus_pos_not_inGene, fivep_minus_pos_not_inGene, 
            threep_plus_pos_inGene, threep_minus_pos_inGene,
            threep_plus_pos_not_inGene, threep_minus_pos_not_inGene)

def getGenePositions(db, chr, strand, txt_db):
    """
    Returns a set that contains every position on the chromsome and strand that
    contains a gene locus.
    """
    pos2gene_set = {}

    gene_select = """SELECT name, start, end
                     FROM gene
                     WHERE chr=\'%s\'
                           AND
                           strand=\'%s\'""" % (chr, strand)

    gene_records = db.getDBRecords_Dict(gene_select, txt_db)

    for gene_row in gene_records:
        gene = gene_row["name"]
        start = int(gene_row["start"])
        end = int(gene_row["end"])

        for i in range(start,end+1):
            updateDictOfSets(pos2gene_set, i, gene)

    return pos2gene_set

def getJunctionSeq(chr_seq, intron_start, intron_end, strand, overhang,
                   rev_comp):

    left_side = chr_seq.extractSeq(intron_start - 1 - overhang + 1,
                                   intron_start - 1)

    right_side = chr_seq.extractSeq(intron_end + 1,
                                    intron_end + 1 + overhang - 1)

    jcn_seq = left_side + right_side

    if rev_comp:
        if strand == "-":
            return jcn_seq.reverseComplement().tostring()

    return jcn_seq.tostring()


def printJcnsFromAnnotatedSS(db, txt_db, chr, chr_seq,
                             outfile, rev_comp, use_unknown_strand,
                             a_file=None, g_file=None, r_file=None,
                             overhang=None):
    """
    Returns a set of joined junctions:
    joined_set_plus
    joined_set_minus

    annotated_fivep_ss_plus
    annotated_fivep_ss_minus
    annotated_threep_ss_plus
    annotated_threep_ss_minus
    """
    joined_set_plus = set([])
    joined_set_minus = set([])


    # Stores {gene: ([fivep_pos,],[threep_pos,])}
    gene2ss_plus = {}
    gene2ss_minus = {}

    # Used to keep track if a splice site is shared by multiple genes
    # {pos:set([gene,])
    fivepss_plus_pos2gene = {}
    fivepss_minus_pos2gene = {}
    threepss_plus_pos2gene = {}
    threepss_minus_pos2gene = {}

    annotated_jcns_plus = set([])
    annotated_jcns_minus = set([])

    # Select all annotated junctions
    select_state = """SELECT DISTINCT gene_name, strand, start, end
                      FROM intron
                      WHERE chr=\'%s\'""" % chr

    records = db.getDBRecords_Dict(select_state, txt_db)

    if records == ():
        print "No genes on %s" % chr
        return (joined_set_plus, joined_set_minus, 
                fivepss_plus_pos2gene, fivepss_minus_pos2gene,
                threepss_plus_pos2gene, threepss_minus_pos2gene)

    for row in records:
        gene = row["gene_name"]

        strand = row["strand"]

        if strand == ".":
            if not use_unknown_strand:
                continue

            # Consider junction in both the plus and minus orientation. And
            # print out the annotated version here.
            intron_start = int(row["start"])
            intron_end = int(row["end"])
            joined_set_plus.add((intron_start, intron_end))        
            joined_set_minus.add((intron_start, intron_end))        
            print_str = "A\t%s%s%s\t%s\t.\t%d\t%d\n" % (gene,
                                                        DELIM,
                                                        gene,
                                                        chr,
                                                        intron_start,
                                                        intron_end)

            if a_file:
                writeToFasta(a_file, print_str, chr_seq, intron_start,
                             intron_end, ".", overhang, rev_comp)

            # Add to plus orientation
            fivep_pos = int(row["start"])
            threep_pos = int(row["end"])

            annotated_jcns_plus.add((fivep_pos, threep_pos))

            if gene in gene2ss_plus:
                gene2ss_plus[gene][0].add(fivep_pos)
                gene2ss_plus[gene][1].add(threep_pos)
            else:
                gene2ss_plus[gene] = (set([fivep_pos]),set([threep_pos]))

            updateDictOfSets(fivepss_plus_pos2gene, fivep_pos, gene)
            updateDictOfSets(threepss_plus_pos2gene, threep_pos, gene)
            
            # Add to minus orientation
            fivep_pos = int(row["end"])
            threep_pos = int(row["start"])

            annotated_jcns_minus.add((threep_pos, fivep_pos))
            
            if gene in gene2ss_minus:
                gene2ss_minus[gene][0].add(fivep_pos)
                gene2ss_minus[gene][1].add(threep_pos)
            else:
                gene2ss_minus[gene] = (set([fivep_pos]),set([threep_pos])) 

            updateDictOfSets(fivepss_minus_pos2gene, fivep_pos, gene)
            updateDictOfSets(threepss_minus_pos2gene, threep_pos, gene)

        elif strand == "-":
            fivep_pos = int(row["end"])
            threep_pos = int(row["start"])

            annotated_jcns_minus.add((threep_pos, fivep_pos))
            
            if gene in gene2ss_minus:
                gene2ss_minus[gene][0].add(fivep_pos)
                gene2ss_minus[gene][1].add(threep_pos)
            else:
                gene2ss_minus[gene] = (set([fivep_pos]),set([threep_pos])) 

            updateDictOfSets(fivepss_minus_pos2gene, fivep_pos, gene)
            updateDictOfSets(threepss_minus_pos2gene, threep_pos, gene)

        else: #strand == "+"
            fivep_pos = int(row["start"])
            threep_pos = int(row["end"])

            annotated_jcns_plus.add((fivep_pos, threep_pos))

            if gene in gene2ss_plus:
                gene2ss_plus[gene][0].add(fivep_pos)
                gene2ss_plus[gene][1].add(threep_pos)
            else:
                gene2ss_plus[gene] = (set([fivep_pos]),set([threep_pos]))

            updateDictOfSets(fivepss_plus_pos2gene, fivep_pos, gene)
            updateDictOfSets(threepss_plus_pos2gene, threep_pos, gene)

    # Now form A, G, and R junctions
    for gene in gene2ss_plus:
        for fivep_pos in gene2ss_plus[gene][0]:
            for threep_pos in gene2ss_plus[gene][1]:
                if (fivep_pos, threep_pos) in joined_set_plus:
                    continue
                joined_set_plus.add((fivep_pos, threep_pos))        

                if fivep_pos < threep_pos:
                    if (fivep_pos, threep_pos) in annotated_jcns_plus:
                        print_str = "A\t%s%s%s\t%s\t+\t%d\t%d\n" % (",".join(list(fivepss_plus_pos2gene[fivep_pos])), 
                                                                    DELIM,
                                                                   ",".join(list(threepss_plus_pos2gene[threep_pos])),
                                                                   chr,
                                                                   fivep_pos,
                                                                   threep_pos)

                        if a_file:
                            writeToFasta(a_file, print_str, chr_seq, fivep_pos,
                                         threep_pos, "+", overhang, rev_comp)

                    else:
                        print_str = "G\t%s%s%s\t%s\t+\t%d\t%d\n" % (",".join(list(fivepss_plus_pos2gene[fivep_pos])), 
                                                                    DELIM,
                                                                   ",".join(list(threepss_plus_pos2gene[threep_pos])),
                                                                   chr,
                                                                   fivep_pos,
                                                                   threep_pos)

                        if g_file:
                            writeToFasta(g_file, print_str, chr_seq, fivep_pos,
                                         threep_pos, "+", overhang, rev_comp)

                        
                else: # fivep_pos >= threep_pos
                    print_str = "R\t%s%s%s\t%s\t+\t%d\t%d\n" % (",".join(list(fivepss_plus_pos2gene[fivep_pos])), 
                                                                DELIM,
                                                               ",".join(list(threepss_plus_pos2gene[threep_pos])),
                                                               chr,
                                                               fivep_pos,
                                                               threep_pos)

                    if r_file:
                        writeToFasta(r_file, print_str, chr_seq, fivep_pos,
                                     threep_pos, "+", overhang, rev_comp)

                
                outfile.write(print_str)
                     
    for gene in gene2ss_minus:
        for fivep_pos in gene2ss_minus[gene][0]:
            for threep_pos in gene2ss_minus[gene][1]:
                if (threep_pos, fivep_pos) in joined_set_minus:
                    continue
                joined_set_minus.add((threep_pos, fivep_pos))        

                if fivep_pos > threep_pos:
                    if (threep_pos, fivep_pos) in annotated_jcns_minus:
                        # The coordinates have to be flipped when printing out
                        # because intron_start > intron_end in the database
                        print_str = "A\t%s%s%s\t%s\t-\t%d\t%d\n" % (",".join(list(fivepss_minus_pos2gene[fivep_pos])),
                                                                    DELIM,
                                                                   ",".join(list(threepss_minus_pos2gene[threep_pos])), 
                                                                   chr,
                                                                   threep_pos, fivep_pos)

                        if a_file:
                            writeToFasta(a_file, print_str, chr_seq, threep_pos,
                                          fivep_pos, "-", overhang, rev_comp)


                    else:
                        print_str = "G\t%s%s%s\t%s\t-\t%d\t%d\n" % (",".join(list(fivepss_minus_pos2gene[fivep_pos])),
                                                                    DELIM,
                                                                   ",".join(list(threepss_minus_pos2gene[threep_pos])), 
                                                                   chr,
                                                                   threep_pos,
                                                                   fivep_pos)

                        if g_file:
                            writeToFasta(g_file, print_str, chr_seq, threep_pos,
                                         fivep_pos, "-", overhang, rev_comp)
            
                else: # fivep_pos <= threep_pos
                    print_str = "R\t%s%s%s\t%s\t-\t%d\t%d\n" % (",".join(list(fivepss_minus_pos2gene[fivep_pos])),
                                                                DELIM,
                                                               ",".join(list(threepss_minus_pos2gene[threep_pos])), 
                                                               chr,
                                                               threep_pos,
                                                               fivep_pos)

                    if r_file:
                        writeToFasta(r_file, print_str, chr_seq, threep_pos,
                                     fivep_pos, "-", overhang, rev_comp)

                outfile.write(print_str)

    return (joined_set_plus, joined_set_minus, 
            fivepss_plus_pos2gene, fivepss_minus_pos2gene,
            threepss_plus_pos2gene, threepss_minus_pos2gene)


def writeToFasta(fasta_file, title_str, chr_seq, intron_start, intron_end, strand,
                 overhang, rev_comp):

    jcn_seq = getJunctionSeq(chr_seq, intron_start, intron_end, strand, 
                             overhang, rev_comp)

    if len(jcn_seq) != 2*overhang:
        return

    title_str = title_str.rstrip("\n")

    (type, genes, chr, strand,
     intron_start_str, intron_end_str) = title_str.split("\t")

    chr = formatChr(chr)
    intron_start = int(intron_start_str)
    intron_end = int(intron_end_str)

    upstrm_end = repr(intron_start - 1)
    upstrm_start = repr(intron_start - overhang)

    dwnstrm_start = repr(intron_end + 1)
    dwnstrm_end = repr(intron_end + overhang)

    gene_comp = genes.split(DELIM)
    genes = ";".join(gene_comp)

    title_comp = [chr, upstrm_start, upstrm_end,
                  dwnstrm_start, dwnstrm_end,
                  genes, strand, type]

    title = DELIM.join(title_comp)
    fasta_file.write(">%s\n%s\n" % (title, jcn_seq))

#################
# END FUNCTIONS #	
#################	
if __name__ == "__main__": main()
